@@ -14,7 +14,7 @@ use namespace::clean -except => 'meta';
 # the double not makes this no longer undef, so exempt from useless constant warnings in older perls
 use constant DEBUG => not not our $DEBUG || $ENV{DATA_VISITOR_DEBUG};
 
-our $VERSION = "0.20";
+our $VERSION = "0.21";
 
 has tied_as_objects => (
 	isa => "Bool",
@@ -69,7 +69,11 @@ sub visit {
 			}
 		}
 
-		push @ret, $self->visit_no_rec_check( $data );
+		if ( defined wantarray ) {
+			push @ret, scalar($self->visit_no_rec_check($data));
+		} else {
+			$self->visit_no_rec_check($data);
+		}
 	}
 
 	return ( @_ == 1 ? $ret[0] : @ret );
@@ -144,48 +148,48 @@ sub visit_hash {
 
 	local $self->{depth} = (($self->{depth}||0) + 1) if DEBUG;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $hash, $hash );
-
-		if ( defined(tied(%$hash)) and $self->tied_as_objects ) {
-			$self->visit_tied(tied(%$hash), $_[1]);
-		} else {
-			$self->visit_hash_entries($_[1]);
-		}
-
-		return;
+	if ( defined(tied(%$hash)) and $self->tied_as_objects ) {
+		return $self->visit_tied_hash(tied(%$hash), $_[1]);
 	} else {
-		if ( defined(tied(%$hash)) and $self->tied_as_objects ) {
-			return $self->visit_tied_hash(tied(%$hash), $_[1]);
-		} else {
-			return $self->visit_normal_hash($_[1]);
-		}
+		return $self->visit_normal_hash($_[1]);
 	}
 }
 
 sub visit_normal_hash {
 	my ( $self, $hash ) = @_;
 
-	my $new_hash = {};
-	$self->_register_mapping( $hash, $new_hash );
+	if ( defined wantarray ) {
+		my $new_hash = {};
+		$self->_register_mapping( $hash, $new_hash );
 
-	%$new_hash = $self->visit_hash_entries($_[1]);
+		%$new_hash = $self->visit_hash_entries($_[1]);
 
-	return $self->retain_magic( $_[1], $new_hash );
+		return $self->retain_magic( $_[1], $new_hash );
+	} else {
+		$self->_register_mapping($hash, $hash);
+		$self->visit_hash_entries($_[1]);
+		return;
+	}
 }
 
 sub visit_tied_hash {
 	my ( $self, $tied, $hash ) = @_;
 
-	my $new_hash = {};
-	$self->_register_mapping( $hash, $new_hash );
+	if ( defined wantarray ) {
+		my $new_hash = {};
+		$self->_register_mapping( $hash, $new_hash );
 
-	if ( blessed(my $new_tied = $self->visit_tied(tied(%$hash), $_[2])) ) {
-		$self->trace( data => tying => var => $new_hash, to => $new_tied ) if DEBUG;
-		tie %$new_hash, 'Tie::ToObject', $new_tied;
-		return $self->retain_magic($_[2], $new_hash);
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_hash, to => $new_tied ) if DEBUG;
+			tie %$new_hash, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_hash);
+		} else {
+			return $self->visit_normal_hash($_[2]);
+		}
 	} else {
-		return $self->visit_normal_hash($_[2]);
+		$self->_register_mapping($hash, $hash);
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
@@ -228,48 +232,50 @@ sub visit_hash_value {
 sub visit_array {
 	my ( $self, $array ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $array, $array );
-
-		if ( defined(tied @$array) and $self->tied_as_objects ) {
-			$self->visit_tied(tied(@$array), $_[1]);
-		} else {
-			$self->visit_array_entries($_[1]);
-		}
-
-		return;
+	if ( defined(tied(@$array)) and $self->tied_as_objects ) {
+		return $self->visit_tied_array(tied(@$array), $_[1]);
 	} else {
-		if ( defined(tied(@$array)) and $self->tied_as_objects ) {
-			return $self->visit_tied_array(tied(@$array), $_[1]);
-		} else {
-			return $self->visit_normal_array($_[1]);
-		}
+		return $self->visit_normal_array($_[1]);
 	}
 }
 
 sub visit_normal_array {
 	my ( $self, $array ) = @_;
 
-	my $new_array = [];
-	$self->_register_mapping( $array, $new_array );
+	if ( defined wantarray ) {
+		my $new_array = [];
+		$self->_register_mapping( $array, $new_array );
 
-	@$new_array = $self->visit_array_entries($_[1]);
+		@$new_array = $self->visit_array_entries($_[1]);
 
-	return $self->retain_magic( $_[1], $new_array );
+		return $self->retain_magic( $_[1], $new_array );
+	} else {
+		$self->_register_mapping( $array, $array );
+		$self->visit_array_entries($_[1]);
+
+		return;
+	}
 }
 
 sub visit_tied_array {
 	my ( $self, $tied, $array ) = @_;
 
-	my $new_array = [];
-	$self->_register_mapping( $array, $new_array );
+	if ( defined wantarray ) {
+		my $new_array = [];
+		$self->_register_mapping( $array, $new_array );
 
-	if ( blessed(my $new_tied = $self->visit_tied(tied(@$array), $_[2])) ) {
-		$self->trace( data => tying => var => $new_array, to => $new_tied ) if DEBUG;
-		tie @$new_array, 'Tie::ToObject', $new_tied;
-		return $self->retain_magic($_[2], $new_array);
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_array, to => $new_tied ) if DEBUG;
+			tie @$new_array, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_array);
+		} else {
+			return $self->visit_normal_array($_[2]);
+		}
 	} else {
-		return $self->visit_normal_array($_[2]);
+		$self->_register_mapping( $array, $array );
+		$self->visit_tied($_[1], $_[2]);
+
+		return;
 	}
 }
 
@@ -291,51 +297,49 @@ sub visit_array_entry {
 sub visit_scalar {
 	my ( $self, $scalar ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $scalar, $scalar );
-
-		if ( defined(tied($$scalar)) and $self->tied_as_objects ) {
-			$self->visit_tied(tied($$scalar), $_[1]);
-		} else {
-			$self->visit($$scalar);
-		}
-
-		return;
+	if ( defined(tied($$scalar)) and $self->tied_as_objects ) {
+		return $self->visit_tied_scalar(tied($$scalar), $_[1]);
 	} else {
-		my $new_scalar;
-		$self->_register_mapping( $scalar, \$new_scalar );
-
-		if ( defined(tied($$scalar)) and $self->tied_as_objects ) {
-			return $self->visit_tied_scalar(tied($$scalar), $_[1]);
-		} else {
-			return $self->visit_normal_scalar($_[1]);
-		}
+		return $self->visit_normal_scalar($_[1]);
 	}
 }
 
 sub visit_normal_scalar {
 	my ( $self, $scalar ) = @_;
 
-	my $new_scalar;
-	$self->_register_mapping( $scalar, \$new_scalar );
+	if ( defined wantarray ) {
+		my $new_scalar;
+		$self->_register_mapping( $scalar, \$new_scalar );
 
-	$new_scalar = $self->visit( $$scalar );
+		$new_scalar = $self->visit( $$scalar );
+
+		return $self->retain_magic($_[1], \$new_scalar);
+	} else {
+		$self->_register_mapping( $scalar, $scalar );
+		$self->visit( $$scalar );
+		return;
+	}
 
-	return $self->retain_magic($_[1], \$new_scalar);
 }
 
 sub visit_tied_scalar {
 	my ( $self, $tied, $scalar ) = @_;
 
-	my $new_scalar;
-	$self->_register_mapping( $scalar, \$new_scalar );
+	if ( defined wantarray ) {
+		my $new_scalar;
+		$self->_register_mapping( $scalar, \$new_scalar );
 
-	if ( blessed(my $new_tied = $self->visit_tied(tied($$scalar), $_[2])) ) {
-		$self->trace( data => tying => var => $new_scalar, to => $new_tied ) if DEBUG;
-		tie $new_scalar, 'Tie::ToObject', $new_tied;
-		return $self->retain_magic($_[2], \$new_scalar);
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_scalar, to => $new_tied ) if DEBUG;
+			tie $new_scalar, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], \$new_scalar);
+		} else {
+			return $self->visit_normal_scalar($_[2]);
+		}
 	} else {
-		return $self->visit_normal_array($_[2]);
+		$self->_register_mapping( $scalar, $scalar );
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
@@ -347,49 +351,49 @@ sub visit_code {
 sub visit_glob {
 	my ( $self, $glob ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $glob, $glob );
-
-		if ( defined(tied(*$glob)) and $self->tied_as_objects ) {
-			$self->visit_tied(tied(*$glob), $_[1]);
-		} else {
-			$self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
-		}
-
-		return;
+	if ( defined(tied(*$glob)) and $self->tied_as_objects ) {
+		return $self->visit_tied_glob(tied(*$glob), $_[1]);
 	} else {
-		if ( defined(tied(*$glob)) and $self->tied_as_objects ) {
-			return $self->visit_tied_glob(tied(*$glob), $_[1]);
-		} else {
-			return $self->visit_normal_glob($_[1]);
-		}
+		return $self->visit_normal_glob($_[1]);
 	}
 }
 
 sub visit_normal_glob {
 	my ( $self, $glob ) = @_;
 
-	my $new_glob = Symbol::gensym();
-	$self->_register_mapping( $glob, $new_glob );
+	if ( defined wantarray ) {
+		my $new_glob = Symbol::gensym();
+		$self->_register_mapping( $glob, $new_glob );
 
-	no warnings 'misc'; # Undefined value assigned to typeglob
-	*$new_glob = $self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
+		no warnings 'misc'; # Undefined value assigned to typeglob
+		*$new_glob = $self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
 
-	return $self->retain_magic($_[1], $new_glob);
+		return $self->retain_magic($_[1], $new_glob);
+	} else {
+		$self->_register_mapping( $glob, $glob );
+		$self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
+		return;
+	}
 }
 
 sub visit_tied_glob {
 	my ( $self, $tied, $glob ) = @_;
 
-	my $new_glob = Symbol::gensym();
-	$self->_register_mapping( $glob, \$new_glob );
+	if ( defined wantarray ) {
+		my $new_glob = Symbol::gensym();
+		$self->_register_mapping( $glob, \$new_glob );
 
-	if ( blessed(my $new_tied = $self->visit_tied($tied, $_[2])) ) {
-		$self->trace( data => tying => var => $new_glob, to => $new_tied ) if DEBUG;
-		tie *$new_glob, 'Tie::ToObject', $new_tied;
-		return $self->retain_magic($_[2], $new_glob);
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_glob, to => $new_tied ) if DEBUG;
+			tie *$new_glob, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_glob);
+		} else {
+			return $self->visit_normal_glob($_[2]);
+		}
 	} else {
-		return $self->visit_normal_glob($_[2]);
+		$self->_register_mapping( $glob, $glob );
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
